# Generated by Django 3.2.15 on 2022-09-06 21:25

from django.db import migrations

from pghistory.models import Context


def install_pgh_attach_context_func(apps, schema_editor):
    Context.install_pgh_attach_context_func(using=schema_editor.connection.alias)


class Migration(migrations.Migration):
    dependencies = [
        ("pghistory", "0003_auto_20201023_1636"),
    ]

    operations = [
        migrations.RunPython(
            install_pgh_attach_context_func, reverse_code=migrations.RunPython.noop
        )
    ]
