# Generated by Django 4.2.6 on 2023-10-14 20:31

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("pghistory", "0005_events_middlewareevents"),
    ]

    operations = [
        migrations.DeleteModel(
            name="AggregateEvent",
        ),
    ]
