{% extends "admin/change_list.html" %}

{% block object-tools-items %}
  	{% if pgh_back %}
    	<li>
      		<a href="{{ pgh_back }}" class="back">Back</a>
    	</li>
  	{% endif %}

  	{{ block.super }}
{% endblock %}